#include <stdint.h>
#include <stdio.h>
#include <stdint.h>
#include <cuda_runtime.h>
#include <string.h>

namespace raylib { 
  #include <raylib.h> 
}


#include "kernel.cuh"
#include "sphere.cuh"
#include "scene.cuh"
#include "camera.cuh"



int main() {
  //bluuuuugh i'll figure out occupancy later, this res are easy
  //calculated manually for gtx1060 with 20 SM, 1024 threads/SM
  const int res_x = 32 * 20;
  const int res_y = 32 * 20;
  const dim3 blockCount(20, 20);
  const dim3 threadCount(32, 32);

  uint8_t *raylib_image;
  //Color texture_data[res_x * res_y];
  raylib::SetTargetFPS(1);


  //see if GPU is connected (my egpu is finicky)
  {
    int temp_device;
    cudaError_t err;
    if((err = cudaGetDevice(&temp_device)) != cudaSuccess) {
      printf("failed to get device!\nError: %s\n", cudaGetErrorString(err));
      return(1);
    }
  }
  
  raylib::SetTraceLogLevel(raylib::LOG_ERROR);
  raylib::InitWindow(res_x, res_y, "cuda teseteroni");
  
  //TODO could probably cut out
  raylib::Image image = raylib::GenImageColor(res_x, res_y, raylib::BLUE);
  raylib::Texture tex = raylib::LoadTextureFromImage(image);


  Sphere *sp1;
  cudaMallocManaged(&sp1, sizeof(Sphere));
  *sp1 = Sphere(make_vect3(0, 0, 1));

  Sphere *sp2;
  cudaMallocManaged(&sp2, sizeof(Sphere));
  *sp2 = Sphere(make_vect3(0, 1, 2));

  Scene *scene;
  cudaMallocManaged(&scene, sizeof(Scene));
  *scene = Scene(); //probably don't need to do?

  scene->add_render_object(sp1);
  scene->add_render_object(sp2);
  
  Camera *camera;
  cudaMallocManaged(&camera, sizeof(Camera));
  *camera = Camera(scene);


  //maybe we shouldn't be infering how to use the raylib::Color type without using
  //the header itself within the kernel (which I honestly don't wanna do)
  cudaMallocManaged(&raylib_image, res_x * res_y * sizeof(raylib::Color));


  while(!raylib::WindowShouldClose()) {
    cudaError_t err;
    //cuda stuff
    render<<<blockCount, threadCount>>>(raylib_image, camera); 
    if((err = cudaGetLastError()) != cudaSuccess) {
      printf("kernel did not launch! Error: %s\n", cudaGetErrorString(err));
    }
    cudaDeviceSynchronize();
    //cudaMemcpy(texture_data, (void **)raylib_image, res_x * res_y * sizeof(Color), cudaMemcpyDeviceToHost);

    raylib::BeginDrawing();
    raylib::ClearBackground(raylib::BLACK);
    raylib::UpdateTexture(tex, raylib_image);
    raylib::DrawTexture(tex, 0, 0, raylib::WHITE);
    raylib::DrawFPS(0, 0);
    raylib::EndDrawing();
  }
  return 0;
}
